cd('/Users/berk/Desktop/Dropbox/Research/MCMC IS/INFORMS JOC Paper/Code')

%% Prelimary Requirements

%Before running this script you will need to make sure that you can

% 1. Have a working installation of CPLEX 12.4. This can be downloaded for 
% free for research purposes through the IBM Academic Initiative at:
% http://www-03.ibm.com/ibm/university/academic/pub/page/software
%

% 2. Be able to compile MEX files in your current version of MATLAB.
% You can learn more about how to compile these files through MATLAB's
% documention using the 'doc mex' command. This is also available at:
% http://www.mathworks.com/help/matlab/ref/mex.html

% 3. If you wish to use MCMC-IS, you will also need the MATLAB KDE Toolbox,
% which can be downloaded for free at:
% http://www.ics.uci.edu/~ihler/code/kde.html
% Note that the MATLAB KDE toolbox also requires you to compile MEX files.

%% Description of the Files

% This script covers three files that are useful for Stochastic Programming:

% 1. mcmcrw - this file is used to generate samples from the zero-variance 
% importance sampling distribution using a Metropolis Hastings algorithm where
% samples are proposed through a random walk process with a fixed step size. 
% Use this file to generate samples if you have a good idea of what the step 
% size for your random walk process.
 
% 2. mcmcam - this file is used to generate samples from the zero-variance 
% importance sampling distribution using an Adaptive Metropolis Hastings 
% algorithm where samples are proposed through a random walk process with
% an automatically selected stepsize. Use this is you do not want to specify
% the step size for your random walk process.
% 
% 3. addcuts - this file is used to solve a linear program with stochastic
% parameters a large number of times. It returns the value of the cut
% parameters, as well as the sample average and sample standard deviation
% of the recourse function over these samples. This file can be used to 
% easily solve SPs using a Monte Carlo method. 
% the samples, 

%% Using The Files to Solve Stochastic Programming Model

% In order to use these files to solve an SP model, you will need to embed
% the MEX files within a sampling-based decomposition algorithm of your choice. 
% The missing 'parts' to actually solve an SP model are simply a forward pass
% procedure to find trial solutions, and a stopping test to compare the values
% of the upper bound and lower bound, and stop the algorithm.
% 
% For generic sample based decomposition procedures, addcuts will suffice.
% If you wish to use MCMC-IS, however, you will first have to:
% 1. generate samples through 'mcmcrw' or 'mcmcam'
% 2. estimate the to generate a set of samples from the zero-variance distribution. 
% 3. generate a second set of samples to use with 'addcuts'

%% Representing a Stochastic Programming Model

% All the files above require a generic way of representing the LPs and 
% random variables associated with a stochastic programming model. In
% this part, we describe how to represent a generic LP with random parameters 
% in MATLAB

% -------------------
%   LP Parameters
% -------------------

% We generally assume that the recourse function in an SP model is an LP
% with:
%
% - N_y variables
% - Meq equality constraints
% - Mineq inequality constraints
%
%In multistage stochastic programming models, each LP is augmented with a 
%cut variable and a set of cutting planes that are used to form a piecewise
%linear representation of the expected future cost function. 
%Thus, our inputs also assume that an LP contains:
%
% - 1 cut variable
% - K cutting planes (used to represent the future cost of each LP)
% - Note that the total number of constraints is M + K.
%
% Thus, our LPs are expressed in the form:
%
% min C'y + theta
% A_eq   * y            =   B_eq_lhs   - E_eq_lhs*x
% A_ineq * y            >=  B_ineq_lhs - E_ineq_lhs*x
% cutcoefs * y + theta  >=  cutlhs
% ub >= y >= lb
% theta > thetalb
%
% where y is the current stage decision, x is the previous stage decision 
% WITHOUT the cut variables 
%
% Given such a model, our files therefore require specifying the following 
%parameters:

% C              is a    N_y x 1 vector
% A              is a    M x N_y matrix    (the first Meq rows are for equality constraints)
% Blhs           is a    M x 1 vector   (the first Meq rows are for equality constraints)
% Elhs           is a    M x N_x matrix (the first Meq rows are for equality constraints)
% ub             is a    N_y x 1 vector
% lb             is a    N_y x 1 vector
% x              is a    N_x x 1 vector (containing the previous stage decisions)
% Meq            is a    1 x 1 integer  (containing the number of equality constraints)
% cutlhs         is a    K x 1 vector (containing the constant in the cuts)
% cutcoefs       is a    K x N_y matrix (containing the coefficients of the cuts)
% thetalb        is a    1 x 1 double (containing the minimum value of theta, the cut variable)

%Note that we have reduced the total number of inputs by concatenating
%the parameters associated with inequality and equality constraints
%so that:
%
% A     = [A_eq;A_ineq]
% Blhs  = [B_eq_lhs; B_ineq_lhs]
% Elhs  = [E_eq_lhs; E_ineq_lhs]

% -------------------
%   Random Variables
% -------------------

%In the LP above, any of the parameters can be random. Let us say that we 
%are solving an LP with D random variables. We indicate the position of
%these D random variables using a D x 2 matrix called 'indices.'

%The first column of indices contains an integer from 1-9 which specifies
%which parameter is random. In particular,
% 
% 1 means the random parameter is in C
% 2 means the random parameter is in Aineq
% 3 means the random parameter is in Aeq
% 4 means the random parameter is in Bineq
% 5 means the random parameter is in Beq
% 6 means the random parameter is in Eineq
% 7 means the random parameter is in Eeq
% 8 means the random parameter is in ub
% 9 means the random parameter is in lb
%
%where Aineq,Aeq,Bineq,Beq,Eineq,Eeq are the smaller matrices that are specified above.
%
%
%The second column of indices contains an integer that specifies the location of this random parameter 
%using a weird C/MEX/MATLAB indexing structure. 
%
%Given a matrix A with M rows and N columns, the index of the i,j-th element is
%
%(i-1) + M*(j-1)
%
%where i is the row index ranging from 1 to M, and j is the column index ranging from 1 to N

%% Representing the Recourse Function of a Newsvendor Model

%% Detailed Description for addcuts

%addcuts is a MEX file that is used to solve a single LP whose parameters
%may contain random variables. It is used to generate the values of cut 
%parameters for cutting plane / decomposition algorithms. It can also be
%used to generate the sample average and sample standard deviation of the
%recourse function. 

%------------------------------------------------------
%RUN
%------------------------------------------------------

%addcuts can be run as

%[alpha beta feasible recourse recoursestd runtime] 
% = 
%addcuts(C,A,Blhs,Elhs,ub,lb,x, Meq,cutlhs,cutcoefs,thetalb,samples,weights,indices)%
%
%
%------------------------------------------------------
%INPUTS
%------------------------------------------------------
%
%
% C              is a    N_y x 1 vector
% A              is a    M x N_y matrix    (the first Meq rows are for equality constraints)
% Blhs           is a    M x 1 vector   (the first Meq rows are for equality constraints)
% Elhs           is a    M x N_x matrix (the first Meq rows are for equality constraints)
% ub             is a    N_y x 1 vector
% lb             is a    N_y x 1 vector
% x              is a    N_x x 1 vector (containing the previous stage decisions)
% Meq            is a    1 x 1 integer  (containing the number of equality constraints)
% cutlhs         is a    K x 1 vector (containing the constant in the cuts)
% cutcoefs       is a    K x N_y matrix (containing the coefficients of the cuts)
% thetalb**      is a    1 x 1 double (containing the minimum value of theta, the cut variable)
% samples**      is a    S x D matrix (containing S samples of a D-dimensional random variables, which can be parameters of the LPs)
% weights        is a    S x 1 vector (containing S weights by which we multiply any MC estimates; this is mainly for importance sampling purposes and can be left blank)
% indices**      is a    D x 2 matrix of INTEGERS that contain the indices of the random LP parameters
% 
% Notes:
%
% - thetalb is a user-defined parameter to capture the lowest value that
%   the recourse function can take. A value of -10^-8 is usually fine.
% 
% - indices has to be an integer data structure, so please make sure to use
%   int32 command to convert it from double to integer)

% - samples is a S x D matrix, which contains S different samples of a 
%   D-dimensional rnadom variable. The COLUMN of the samples matrix should 
%   be in the SAME ORDER as the ROWS of the indices matrix so the first row 
%   of indices corresponds to the same random variable as the first column 
%   of the indices matrix etc.
%
%------------------------------------------------------
%OUTPUTS
%------------------------------------------------------

%The outputs from addcuts are
%
%[alpha beta feasible recourse recoursestd runtime]
%
%alpha and beta are cut variables are that added to the previous stage LP. 
%for reference, the cut formulation that I use is
%
%beta * x >= alpha
%
%feasible is a TRUE/FALSE value telling you whether or not the recourse function is feasible 
%(note, I have not implemented feasibility cuts)
%
%recourse is a value of the recourse
%recoursestd is the value of the standard deviation of the recourse
%runtime is the CPLEX runtime

%% Test Run for addcuts


%compile addcuts
clc
clear all
close all
mex -O -I/Applications/IBM/ILOG/CPLEX_Studio_Academic124/cplex/include/ ...
    -L/Applications/IBM/ILOG/CPLEX_Studio_Academic124/cplex/lib/x86-64_darwin9_gcc4.0/static_pic ...
     LDFLAGS='\$LDFLAGS -framework CoreFoundation -framework IOKit'...
    -lcplex ...
    addcuts.c 

clc
clear all
close all
load('addcuts_test_data.mat')

[alpha beta feasible recourse recoursestd runtime] = addcuts(C,A,Blhs,Elhs,ub,lb,x, Meq,cutlhs,cutcoefs,thetalb,samples,weights,indices)%

%% Detailed Description for mcmcrw

%------------------------------------------------------
%RUN
%------------------------------------------------------

%addcuts can be run as

%------------------------------------------------------
%INPUTS
%------------------------------------------------------


%mexpdfs: D x 1 vector of CODES describe the type of PDF that will be used
%mexscales: D x 1 vector for the scaling factor for the PDFs
%mexsigmas: D x 1 vector for the step size of the MCMC Chain
%mexstart: D x 1 vector for the starting point of the MCMC chain
%mexsteps: N_t x D matrix of N(0,1) samples which represent the steps that are taken for each accept/reject evaluation in the MCMC Chain 
%mexacceptprobs: N_t x 1  vector of U(0,1) where N_t = total number of accept reject evaulations in the MCMC Chain
%nburnin_mex : # of burnin samples for the MCMC Chain
%nrequired_mex: # of accepted samples to produce for the MCMC Chain

%Note: for each random variable, we assume that it has the form:
%
%
% a * f(X)
%
%where a is the scaling factor we put in mexscales
%
%and
%
%f(X) = exp(X) if mexpdfs = 0 for this variable
%
%f(x) = exp( 0.5 * pow(x, 2.0) - 0.125* pow(x - 3.0, 2)) + exp(0.5 * pow(x, 2.0) - 0.125*pow(x + 1.0, 2.0)));
%if mexpdfs = 1
% the latter corresponds to the multimodal density
%
%
%function [mexacceptprobs, mexsteps] = mexrandfunction(n_mex)
%            mexacceptprobs = %n_mex x 1 vector of U(0,1) - i.e. rand(n_mex,1)
%            mexsteps  = %n_mex x D vector of N(0,1) - i.e. randn(n_mex, D)
%end
%
% the function handle to refer to the actual function
% randfun_in = @(n_mex) mexrandfunction(n_mex)

%
%to add more PDF types, you need only add another function to lines 64 - 85

% This function requires U(0,1) and  N(0,1) variables described AHEAD of time
% to run the Random Walk process in the MEX environment. 
%
% Right now, since we do not know how many accept/reject trials are required ahead of time, 
% we provide N_t of these values to the MEX file and if it runs out of these values then 
% the program is designed to call MATLAB to generate N_t more samples.
%
% Note that this only, happens when, for instance, more than N_t accept/reject trials 
% are required to generate nrequired_mex samples). As such it is useful to have a good 
% idea of what the acceptance rate is before you run this function, and to make N_t 
% as large as possible (N_t should = n_required_mex / acceptance_rate of MCMC algorithm)
%
%------------------------------------------------------
%OUTPUTS
%------------------------------------------------------

%% Test Run for mcmcrw


%compile addcuts
clc
clear all
close all
mex -O -I/Applications/IBM/ILOG/CPLEX_Studio_Academic124/cplex/include/ ...
    -L/Applications/IBM/ILOG/CPLEX_Studio_Academic124/cplex/lib/x86-64_darwin9_gcc4.0/static_pic ...
     LDFLAGS='\$LDFLAGS -framework CoreFoundation -framework IOKit'...
    -lcplex ...
    mcmcrw.c 

load('mcmcrw_test_data.mat');



%% Detailed Description for mcmcam

%% Test Run for mcmcam
