/* Set Required File Headers */
#include "mex.h"
#include "matrix.h"
#include <stdlib.h>
#include <string.h>
#include "ilcplex/cplex.h"
#include "math.h"
#include <pthread.h>

/* Define Input Variable Positions */
#define C_IN        0
#define A_IN        1
#define B_IN        2
#define E_IN        3
#define UB_IN       4
#define LB_IN       5
#define X_IN        6
#define MEQ_IN      7
#define CUTLHS_IN   8
#define CUTCOEFS_IN 9
#define THETALB_IN  10
#define SAMPLES_IN  11
#define WEIGHTS_IN  12
#define INDICES_IN  13

/* Define Output Variable Positions */
#define ALPHA_OUT   0
#define BETA_OUT    1
#define FEAS_OUT    2
#define REC_OUT     3
#define RECSTD_OUT  4
#define TIME_OUT    5

/* Define Machine Precision */
#define EPSILON     0.0000000000001
#define MAXNUM      100000000000000

/* Declare Static Variables */
static   pthread_mutex_t openCPLEXLock = PTHREAD_MUTEX_INITIALIZER;

int fequal(double a, double b)
{
    return ( fabs(a-b) < EPSILON );
}

/* Mex Function */
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    /*                                                                       */
    /*                           Initialize Variables                        */
    /*                                                                       */
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    
    /* Check Number of Input and Output Arguments */
    if      (nrhs!=14) { mexErrMsgIdAndTxt( "invalidNumInputs", "14 inputs required: C, A, Blhs, Elhs, ub, lb, x, Meq, cutlhs, cutcoefs, thetalb, samples, weights, indices"); }
    else if (nlhs!=6)  { mexErrMsgIdAndTxt( "invalidNumInputs", "6 outputs required: alpha, beta, feasible, recourse, recoursestd ");}
    
    /* Declare Reference Variables for Loops */
    int                     i, j, s;
    
    /* Declare and Obtain Reference Variables for LP Size */
    int     N                   = (int) mxGetM(prhs[C_IN]);
    int     Mt                  = (int) mxGetM(prhs[A_IN]);
    int     Meq                 = (int) *mxGetPr(prhs[MEQ_IN]);
    int     D                   = (int) mxGetN(prhs[SAMPLES_IN]);
    int     S                   = (int) mxGetM(prhs[SAMPLES_IN]);
    int     Nx                  = (int) mxGetM(prhs[X_IN]);
    int     Mcuts               = (int) mxGetM(prhs[CUTLHS_IN]);
    int     WEIGHTED            = (int) ( mxGetM(prhs[WEIGHTS_IN]) > 0 );
    int     Mineq               = Mt - Meq;
    int     Nt                  = N + (Mcuts > 0);
    double  SS                  = (double) S;
    
    /* Create Pointers for Input Variables */
    double  *C              = mxGetPr(prhs[C_IN]);
    double  *A              = mxGetPr(prhs[A_IN]);
    double  *B              = mxGetPr(prhs[B_IN]);
    double  *E              = mxGetPr(prhs[E_IN]);
    double  *ub             = mxGetPr(prhs[UB_IN]);
    double  *lb             = mxGetPr(prhs[LB_IN]);
    double  *x              = mxGetPr(prhs[X_IN]);
    double  *cutlhs         = mxGetPr(prhs[CUTLHS_IN]);
    double  *cutcoefs       = mxGetPr(prhs[CUTCOEFS_IN]);
    double  *thetalb        = mxGetPr(prhs[THETALB_IN]);
    double  *samples        = mxGetPr(prhs[SAMPLES_IN]);
    double  *weights        = mxGetPr(prhs[WEIGHTS_IN]);
    int     *indices        = (int *) mxGetData(prhs[INDICES_IN]);
    
    /*Initialize CPLEX Environment */
    CPXENVptr       env     = NULL;
    CPXLPptr        lp      = NULL;
    int             status  = 0;
    int             lpstat  = 0;
    
    /*Declare Arrays Used to Build LPs in CPLEX */
    double          *lhs;
    double          *rmatval;
    int				*rmatind;
    int				*rmatbeg;
    char            *rmatsense;
    int				nnz = 0;
    
    /* Declare Arrays Used to Build Cut Variables in CPLEX */
    double          ctheta = 1, thetaub = CPX_INFBOUND;
    double          *cmatval;
    int				*cmatind;
    int				*cmatbeg;
    char            *cmatsense;
    int				nnzcuts = 0;
    
    /* Declare Variables and Pointers Used to Change Random LP Parameters*/
    int           s_obj_cnt = 0,    s_ub_cnt = 0,		 s_lb_cnt = 0,	 s_coef_cnt = 0;
    int          *s_obj_ind,        *s_ub_ind,			*s_lb_ind,		*s_coef_rows,    *s_coef_cols,    *s_lhs_ind;
    double       *s_obj_vals,       *s_ub_vals,         *s_lb_vals,		*s_coef_vals;
    char                            *s_ub_lu,			*s_lb_lu;
    
    /* Count Number of RVs in Each LP Parameter */
    for (i = 0; i < D; i ++ ){
        
        switch ( indices[i] ) {
            /*C*/   case 1:  s_obj_cnt++  ; break;
            /*A*/   case 2:  s_coef_cnt++ ; break;
            /*Aeq*/ case 3:  s_coef_cnt++ ; break;
            /*ub*/  case 8:  s_ub_cnt++   ; break;
            /*lb*/  case 9:  s_lb_cnt++   ; break;
        }
    }
    
    /* Flag LP  Parameters Types Containing RVs */
    int CHGOBJ  = ( s_obj_cnt > 0);
    int CHGCOEF = ( s_coef_cnt > 0);
    int CHGLHS  = ( D - s_obj_cnt -s_coef_cnt -s_ub_cnt -s_lb_cnt > 0 );
    int CHGUB   = ( s_ub_cnt > 0);
    int CHGLB   = ( s_lb_cnt > 0);
    
    /* Allocate Space and Initialize RV Variables for C	*/
    if ( CHGOBJ ) {
        
        s_obj_ind        = (int *)      malloc (s_obj_cnt   * sizeof(int));
        s_obj_vals       = (double *)   malloc (s_obj_cnt   * sizeof(double));
        
        j = 0;
        for (i = 0; i < D; i++) {
            if ( indices[i] == 1) {
                s_obj_ind[j] = indices[i+D];
                j ++;
            }
        }
    }
    
    /* Allocate Space and Initialize RV Variables for A and Aeq */
    if ( CHGCOEF ) {
        
        s_coef_vals      = (double *)   malloc (s_coef_cnt  * sizeof(double));
        s_coef_cols      = (int *)      malloc (s_coef_cnt  * sizeof(int));
        s_coef_rows      = (int *)      malloc (s_coef_cnt  * sizeof(int));
        
        j = 0;
        for (i = 0; i < D; i++) {
            if ( indices[i] == 2 || indices[i] == 3) {
                s_coef_rows[j] = (int) indices[i+D] % Mt;
                s_coef_cols[j] = (int) (indices[i+D] - s_coef_rows[j]) / Mt;
                j ++;
            }
        }
    }
    
    /* Allocate Space and Initialize RV Variables for lhs vector ( B, Beq, E, Eq) */
    if ( CHGLHS ) {
        s_lhs_ind          = (int *) malloc (Mt * sizeof(int));
        for (i = 0; i < Mt; i++) {
            s_lhs_ind[i] = i;
        }
    }
    
    /* Allocate Space and Initialize RV Variables for ub */
    if ( CHGUB ) {
        s_ub_ind         = (int *)      malloc (s_ub_cnt    * sizeof(int));
        s_ub_vals        = (double *)   malloc (s_ub_cnt    * sizeof(double));
        s_ub_lu          = (char *)     malloc (s_ub_cnt    * sizeof(char));
        
        j = 0;
        for (i = 0; i < D; i++) {
            if ( indices[i] == 1) {
                s_ub_ind[j] = indices[i+D];
                s_ub_lu[j]  = 'G';
                j ++;
            }
        }
    }
    
    /* Allocate Space and Initialize RV Variables for lb */
    if ( CHGLB ) {
        s_lb_ind         = (int *)      malloc (s_lb_cnt    * sizeof(int));
        s_lb_lu          = (char *)     malloc (s_lb_cnt    * sizeof(char));
        s_lb_vals        = (double *)   malloc (s_lb_cnt    * sizeof(double));
        
        j = 0;
        for (i = 0; i < D; i++) {
            if ( indices[i] == 1) {
                s_lb_ind[j] = indices[i+D];
                s_lb_lu[j]  = 'L';
                j ++;
            }
        }
    }
    
    
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    /*                                                                       */
    /*                           Setup LP Parameters                         */
    /*                                                                       */
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
  
    /*Open CPLEX Environment */
    status = 1;
    while ( status ) {
        pthread_mutex_lock(&openCPLEXLock);
        env = CPXopenCPLEX (&status);
        pthread_mutex_unlock(&openCPLEXLock);
    }
    
    if ( env == NULL ) {
        mexPrintf("%s", "Could not open CPLEX environment.\n");
        goto TERMINATE;
    }
    
    /*Create CPLEX LP Object */
    lp = CPXcreateprob (env, &status, "LP");
    if ( lp == NULL ) {
        mexPrintf("%s", "Failed to create LP.\n");
        goto TERMINATE;
    }
    
    /* Allocate Memory According to Array Sizes */
    lhs             = (double *)    malloc (Mt          * sizeof(double));
    rmatval         = (double *)    malloc (Mt * N      * sizeof(double));
    rmatind         = (int *)       malloc (Mt * N      * sizeof(int));
    rmatbeg         = (int *)       malloc (Mt          * sizeof(int));
    rmatsense       = (char *)      malloc (Mt          * sizeof(char));
    
    /* Create Sense Vector for Constraints */
    for (i = 0;   i < Meq;  i++) { rmatsense[i] = 'E';} /*Equality Constraints*/
    for (i = Meq; i < Mt;   i++) { rmatsense[i] = 'G';} /*Inequality Constraints*/
    
    /* Add LP Objective */
    CPXchgobjsen (env, lp, CPX_MIN);
    status = CPXnewcols(env, lp, N, C, lb, ub, NULL, NULL);
    
    /* Add Equality and Inequality Constraints */
    for (i = 0; i < Mt; i++) {
        rmatbeg[i] = nnz;
        for (j = 0; j < N; j++) {
            rmatind[nnz] = j;
            rmatval[nnz] = A[i + j * Mt];
            nnz ++;
        }
    }
    
    /* Calculate Constraint LHS */
    for (i = 0; i < Mt; i++) {
        lhs[i] = B[i];
        for (j = 0; j < Nx; j++) {
            lhs[i] = lhs[i] - E[i + j * Mt] * x[j];
        }
    }
    
    status = CPXaddrows(env, lp, 0, Mt, nnz, lhs, rmatsense, rmatbeg, rmatind, rmatval, NULL, NULL);
    
    /* Free Memory */
    free(rmatbeg);
    free(rmatind);
    free(rmatval);
    free(rmatsense);
    
    /* Add Cut Variable and Cut Constraints */
    if (Mcuts > 0) {
        
        cmatsense       = (char *)   malloc (Mcuts          * sizeof(char));
        cmatval         = (double *) malloc (Mcuts * Nt     * sizeof(double));
        cmatind         = (int *)    malloc (Mcuts * Nt     * sizeof(int));
        cmatbeg         = (int *)    malloc (Mcuts			* sizeof(int));
        
        for (i = 0; i < Mcuts; i++)
        {
            cmatsense[i] = 'G';
            cmatbeg[i]   = nnzcuts;
            for (j = 0; j < Nt; j++) {
                cmatind[nnzcuts] = j;
                cmatval[nnzcuts] = cutcoefs[i + j * Mcuts];
                nnzcuts++;
            }
        }
        
        /* Add Cut Variable */
        status = CPXaddcols (env, lp, 1, 0, &ctheta, NULL, NULL, NULL, thetalb, &thetaub, NULL);
         
        /* Add Cut Constraints */
        status = CPXaddrows (env, lp, 0, Mcuts, nnzcuts, cutlhs, cmatsense, cmatbeg, cmatind, cmatval, NULL, NULL);
        
        /* Free Memory */
        free(cmatbeg);
        free(cmatind);
        free(cmatval);
        free(cmatsense);     
        
    }

    /* Determine If We Can Skip Bounds and Reduced Costs Before Calculating Cuts */
    int     SKIPLB = 0;
    int     SKIPUB = 0;
    
    if ( !CHGLB ) {
        for (i = 0;  i < N;  i++){
            if ( fequal(lb[i],0.0) ) {
                SKIPLB = 1;
                break;
            }
        }
    }
    
    if ( !CHGUB ) {
        for (i = 0; i < N;  i++){
            if ( isinf(ub[i]) ) {
                SKIPUB = 1;
                break;
            }
        }
    }
    
    int SKIPBDS = (SKIPLB && SKIPUB && (Mcuts < 1));
    
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    /*                                                                       */
    /*             Solve LP and Calculate Sample Average Cut                 */
    /*                                                                       */
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    
    /* Initialize Output Variables */
    plhs[ALPHA_OUT]         = mxCreateDoubleScalar(0.0);
    plhs[BETA_OUT]          = mxCreateDoubleMatrix((mwSize) 1, (mwSize) Nx, mxREAL);
    plhs[FEAS_OUT]          = mxCreateDoubleScalar(1.0);
    plhs[TIME_OUT]          = mxCreateDoubleScalar(0.0);
    
    /* Create Pointers to Output Variables*/
    double *alpha           = mxGetPr(plhs[ALPHA_OUT]);
    double *beta            = mxGetPr(plhs[BETA_OUT]);
    double *feasible        = mxGetPr(plhs[FEAS_OUT]);
    double *runtime         = mxGetPr(plhs[TIME_OUT]);
    
    /* Initialize Standard Weighting */
    double w                = 1.0 / SS;
    
    /* Declare Pointers and Allocate Memory for Solution Variables */
    double          *objvals, *duals, *reducedcosts;
    double			beforetime, aftertime;
    int             *colstat;
    
    objvals         = (double *)    malloc (S              * sizeof(double));
    duals           = (double *)    malloc ((Mt + Mcuts)   * sizeof(double));
    
    /* Set CPLEX Parameters */
    CPXsetintparam(env, CPX_PARAM_THREADS, 1);
    
    /* Obtain Recourse, Solution, Duals, Slacks, Reduced Costs and Basis Status */
    if ( SKIPBDS ) {
        
        /* Cut Loop */
        for (s = 0; s < S; s++) {
            
            /* Obtain New Sample Values for Random LP Parameters*/
            j = 0;
            for (i = 0; i < D; i ++) {
                switch (indices[i]) {
                    /*C*/   case 1: s_obj_vals[j]   = samples[s + i*S]; if (s_obj_cnt   > j + 1){j++;}  else { j = 0;} break;
                    /*A*/   case 2: s_coef_vals[j]  = samples[s + i*S]; if (s_coef_cnt  > j + 1){j++;} else { j = 0;} break;
                    /*Aeq*/ case 3: s_coef_vals[j]  = samples[s + i*S]; if (s_coef_cnt  > j + 1){j++;} else { j = 0;} break;
                    /*B*/   case 4: B[indices[i+D]] = samples[s + i*S];                                              break;
                    /*Beq*/ case 5: B[indices[i+D]] = samples[s + i*S];                                              break;
                    /*E*/   case 6: E[indices[i+D]] = samples[s + i*S];                                              break;
                    /*Eeq*/ case 7: E[indices[i+D]] = samples[s + i*S];                                              break;
                }
            }
            
            /* Change LP Parameters in CPLEX LP to New Sample Values */
            if ( CHGOBJ ) { status = CPXchgobj(env, lp, s_obj_cnt, s_obj_ind, s_obj_vals);}
            if ( CHGCOEF ){ status = CPXchgcoeflist(env, lp, s_coef_cnt, s_coef_rows, s_coef_cols, s_coef_vals); }
            if ( CHGLHS ) {
                for (i = 0; i < Mt; i++) {
                    lhs[i] = B[i];
                    for (j = 0; j < Nx; j++) { lhs[i] = lhs[i] - E[i + j * Mt] * x[j]; }
                }
                status = CPXchgrhs(env, lp, Mt, s_lhs_ind, lhs);
            }
            
            /* Optimize LP */
            status = CPXgettime(env, &beforetime);
            status = CPXlpopt(env, lp);
            status = CPXgettime(env, &aftertime);
            *runtime = *runtime + (aftertime - beforetime);
            
            /* Obtain Objective Value and Duals */
            status = CPXsolution (env, lp, &lpstat, &objvals[s], NULL, duals, NULL, NULL);
            
            if ( lpstat == CPX_STAT_INFEASIBLE ){
                *feasible = 0;
            }
            
            if ( duals == NULL ) {
                mexPrintf("%s\n", "Could not allocate memory for solution.");
                goto TERMINATE;
            }
            
            /* Update Weight */
            if ( WEIGHTED ) {
                w = weights[s] / SS;
                objvals[s] =  weights[s] * objvals[s];
            }
            
            /* Calculate Alpha */
            for (i = 0; i < Mt; i++)    { *alpha = *alpha + w * duals[i] * B[i]; }
            
            /* Calculate Beta */
            for (i = 0; i < Nx; i++) {
                for (j = 0; j < Mt; j++) {
                    beta[i] = beta[i] + w * duals[j] * E[j + i * Mt];
                }
            }
            
        }
        
    } else {
        
        /* Allocate Space + Initialize for Reduced Costs */
        reducedcosts    = (double *)    malloc (Nt             * sizeof(double));
        colstat         = (int *)       malloc (Nt             * sizeof(int));
        
        for (s = 0; s < S; s++) {
            
            /* Obtain New Sample Values for Random LP Parameters*/
            j = 0;
            for (i = 0; i < D; i ++) {
                switch (indices[i]) {
                    /*C*/   case 1: s_obj_vals[j]   = samples[s + i*S]; if (s_obj_cnt > j + 1){j++;}  else { j = 0;} break;
                    /*A*/   case 2: s_coef_vals[j]  = samples[s + i*S]; if (s_coef_cnt > j + 1){j++;} else { j = 0;} break;
                    /*Aeq*/ case 3: s_coef_vals[j]  = samples[s + i*S]; if (s_coef_cnt > j + 1){j++;} else { j = 0;} break;
                    /*B*/   case 4: B[indices[i+D]] = samples[s + i*S];                                              break;
                    /*Beq*/ case 5: B[indices[i+D]] = samples[s + i*S];                                              break;
                    /*E*/   case 6: E[indices[i+D]] = samples[s + i*S];                                              break;
                    /*Eeq*/ case 7: E[indices[i+D]] = samples[s + i*S];                                              break;
                    /*ub*/  case 8: s_ub_vals[j]    = samples[s + i*S]; if (s_ub_cnt > j + 1){j++;}  else { j = 0;}  break;
                    /*lb*/  case 9: s_lb_vals[j]    = samples[s + i*S]; if (s_lb_cnt > j + 1){j++;}  else { j = 0;}  break;
                }
            }
            
            /* Change LP Parameters in CPLEX LP to New Sample Values */
            if ( CHGOBJ ) { status = CPXchgobj(env, lp, s_obj_cnt, s_obj_ind, s_obj_vals);}
            if ( CHGCOEF ){ status = CPXchgcoeflist(env, lp, s_coef_cnt, s_coef_rows, s_coef_cols, s_coef_vals); }
            if ( CHGUB )  { status = CPXchgbds(env, lp, s_ub_cnt, s_ub_ind, s_ub_lu, s_ub_vals);}
            if ( CHGLB )  { status = CPXchgbds(env, lp, s_lb_cnt, s_lb_ind, s_lb_lu, s_lb_vals);}
            if ( CHGLHS ) {
                for (i = 0; i < Mt; i++) {
                    lhs[i] = B[i];
                    for (j = 0; j < Nx; j++) { lhs[i] = lhs[i] - E[i + j * Mt] * x[j]; }
                }
                status = CPXchgrhs(env, lp, Mt, s_lhs_ind, lhs);
            }
            
            /* Optimize LP */
            status = CPXgettime(env, &beforetime);
            status = CPXlpopt(env, lp);
            status = CPXgettime(env, &aftertime);
            *runtime = *runtime + (aftertime - beforetime);
            
            /* Obtain Objective Value and Duals */
            status = CPXsolution (env, lp, &lpstat, &objvals[s], NULL, duals, NULL, reducedcosts);
            status = CPXgetbase(env, lp, colstat, NULL);
            
            if ( lpstat == CPX_STAT_INFEASIBLE ){
                *feasible = 0;
            }
            
            if ( duals == NULL || reducedcosts == NULL ) {
                status = CPXERR_NO_MEMORY;
                mexPrintf("%s\n", "Could not allocate memory for solution.");
                goto TERMINATE;
            }
            
            
            /* Update Weight */
            if ( WEIGHTED ) {
                w = weights[s] / SS;
                objvals[s] =  weights[s] * objvals[s];
            }
            
            /* Calculate Alpha */
            for (i = 0; i < Mt;     i++) { *alpha = *alpha + w * duals[i] * B[i]; }
            for (i = 0; i < Mcuts;  i++) { *alpha = *alpha + w * duals[Mt + i] * cutlhs[i]; }
            for (i = 0; i < N;      i++) {
                if ( !isnan(reducedcosts[i]) ) {
                    if      (colstat[i] == CPX_AT_UPPER) { *alpha = *alpha + w * reducedcosts[i] * ub[i]; } /* Var at UB */
                    else if (colstat[i] == CPX_AT_LOWER) { *alpha = *alpha + w * reducedcosts[i] * lb[i]; } /* Var at LB */
                }
            }
            
            /* Add Extra Cut Variable Thing */
            if ( (Mcuts > 0) && (colstat[N] == CPX_AT_LOWER) ){
                if ( !isnan(reducedcosts[N]) && !fequal(reducedcosts[N], 0.0) ) {
                    *alpha = *alpha + w * reducedcosts[N] * *thetalb;
                }
            }
            
            /* Calculate Beta */
            for (i = 0; i < Nx; i++) {
                for (j = 0; j < Mt; j++) {
                    beta[i] = beta[i] + w * duals[j] * E[j + i * Mt];
                }
            }
        }
        
        /* Free colstat and reduced costs */
        free(colstat);
        free(reducedcosts);
    }
    
    /* Free Memory */
    free(duals);
    free(lhs);
    
    
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    /*                                                                       */
    /*                     Calculate Recourse Statistics                     */
    /*                                                                       */
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    
    /* Obtain Pointer to Output Statistics */
    plhs[REC_OUT]           = mxCreateDoubleScalar(0.0);
    plhs[RECSTD_OUT]        = mxCreateDoubleScalar(0.0);
    
    /*Set Output Statistic Pointers */
    double *recourse        = mxGetPr(plhs[REC_OUT]);
    double *recoursestd     = mxGetPr(plhs[RECSTD_OUT]);
    
    /* Calculate Recourse / Recourse Sample St. Dev if Problem Was Feasible */
    if ( *feasible ) {
        
        /* Calculate Mean */
        for (s = 0; s < S; s++){ *recourse = *recourse + objvals[s] / SS; };

        
        /* Calculate Recourse Sample St. Dev */
        if (S > 1){
         for (s = 0; s < S; s++){ *recoursestd = *recoursestd  + pow(*recourse - objvals[s], 2.0) / (SS - 1.0); }
             *recoursestd = sqrt(*recoursestd);
             }

    }
    
    /* Free Memory */
    free(objvals);
    
    goto TERMINATE;
    
    
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    /*                                                                       */
    /*                     Termination  Procedure                            */
    /*                                                                       */
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    
    
    /* Termination Procedures */
    TERMINATE:
        
        /* Free up the problem as allocated by CPXcreateprob, if necessary */
        if ( lp != NULL ) {
            status = CPXfreeprob (env, &lp);
            if ( status ) { mexPrintf("%s%d\n", "CPXfreeprob failed, error code :", status); }
        }
        
        /* Free up the CPLEX environment, if necessary */
        if ( env != NULL ) {
            status = CPXcloseCPLEX (&env);
            if ( status ) {
                mexPrintf("%s\n", "Could not close CPLEX environment.");
            }
        }
        
        if ( CHGOBJ ) {
            
            if (s_obj_vals != NULL){
                free(s_obj_vals);
            }
            
            if (s_obj_ind != NULL){
                free(s_obj_ind);
            }
        }
        
        if ( CHGUB ) {
            
            if (s_ub_lu != NULL){
                free(s_ub_lu);
            }
            
            if (s_ub_ind != NULL){
                free(s_ub_ind);
            }
            
            if (s_ub_vals != NULL){
                free(s_ub_vals);
            }
            
        }
        
        if ( CHGLB ) {
            
            if (s_lb_lu != NULL){
                free(s_lb_lu);
            }
            
            if (s_lb_ind != NULL){
                free(s_lb_ind);
            }
            
            if (s_lb_vals != NULL){
                free(s_lb_vals);
            }
        }
        
        if ( CHGCOEF ) {
            
            if (s_coef_rows != NULL){
                free(s_coef_rows);
            }
            
            if (s_coef_cols != NULL){
                free(s_coef_cols);
            }
            
            if (s_coef_vals != NULL){
                free(s_coef_vals);
            }
        }
        
        if ( CHGLHS ) {
            if (s_lhs_ind != NULL){
                free(s_lhs_ind);
            }
        }
        
}