/* Set Required File Headers */
#include "mex.h"
#include "matrix.h"
#include <stdlib.h>
#include <string.h>
#include "ilcplex/cplex.h"
#include "math.h"
#include <pthread.h>

/* Define Input Variable Positions */
#define C_IN            0
#define A_IN            1
#define B_IN            2
#define E_IN            3
#define UB_IN           4
#define LB_IN           5
#define X_IN            6
#define MEQ_IN          7
#define CUTLHS_IN       8
#define CUTCOEFS_IN     9
#define THETALB_IN      10
#define INDICES_IN      11
#define PDF_IN          12
#define SCALES_IN       13
#define SIGMAS_IN       14
#define START_IN        15
#define RANDFUN_IN      16
#define CHOLFUN_IN      17
#define STEPFACTOR_IN   18
#define EPSILON_IN      19
#define NBURNIN_IN      20
#define NADAPTIVE_IN    21
#define NREQUIRED_IN    22
#define NMAX_IN         23
#define MIN_ACCEPT_IN   24
#define MIN_S_IN        25

/* Define Output Variable Positions */
#define ACC_SAMPLES_OUT     0
#define ACC_COUNTS_OUT      1
#define NACCEPT_OUT         2
#define NTOTAL_OUT          3
#define FORCEDOUT_OUT       4
#define TIME_OUT            5

/* Constants */
#define NORM_CONS           0.398942280401433 /* = 1/sqrt(2*pi) */

/* Declare Static Variables */
static   pthread_mutex_t openCPLEXLock = PTHREAD_MUTEX_INITIALIZER;

double mvnpdf(double x[], double sigmas[], int D)
{
    int    i;
    double sum = 0.0;
    double prod = 1.0;
    double DD = (double) D;
    
    for (i = 0; i < D; i++){
        sum = sum + pow(x[i]/sigmas[i],2.0);
        prod = prod*sigmas[i];
    }       
            
    return(pow(NORM_CONS,DD)*(1.0/prod)*exp(-sum/2));
}

/* LN / MM Transform Function */
double strans(double x, double scale, int pdf)
{
    double sample = 0.0;
    
    switch ( pdf ) {
        
        /*LN*/
        case 0:
            
            sample = scale * exp(x);
            break;
            
            /*MM*/
        case 1:
            
            sample = scale * (exp( 0.5 * pow(x, 2.0) - 0.125* pow(x - 3.0, 2)) + exp(0.5 * pow(x, 2.0) - 0.125*pow(x + 1.0, 2.0)));
            break;
    }
    
    return sample;
}


void updateCovariance(double *Cov, double *newdata, double *mean, double *newmean, double stepfactor, double epsilon, int n, int s){
    
    /* Note that s is NOT zero-indexed -> s = 1 means 1 sample etc.*/
    int i, j;
    double ss = (double) s;
    double sd = pow(stepfactor, 2.0);

    /* Calculate New Mean*/
    for (i = 0; i < n; i++){ newmean[i] = ((ss-1)*mean[i] + newdata[i])/ ss; }
    
    /* Calculate Covariance*/
    for (j = 0; j < n; j++) {
        
        /* for i = j */
        Cov[j+n*j] = (ss-1.0)*Cov[j+n*j] + sd*(ss*pow(mean[j],2.0) -(ss+1)*pow(newmean[j],2.0) + pow(newdata[j],2.0) + epsilon);
        Cov[j+n*j] = Cov[j+n*j]/ss;
        
        /* for i > j */
        for (i = j + 1; i < n; i ++) {
            
            /* Recursively Calculate Covariance Matrix */
            Cov[i+n*j] = (ss-1.0)*Cov[i+n*j] +  sd*( ss*mean[i]*mean[j] -(ss+1.0)*newmean[i]*newmean[j] + newdata[i]*newdata[j]);
            Cov[i+n*j] = Cov[i+n*j]/ss;
            
            /* Exploit Symmetry in Covariance Matrix */
            Cov[j+n*i] = Cov[i+n*j];
        }
    }
    
    /*Update Mean*/
    for (i = 0; i < n; i++){ mean[i] = newmean[i]; }
}


/* Mex Function */
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    /*                                                                       */
    /*                           Initialize Variables                        */
    /*                                                                       */
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    
    /* Check Number of Input and Output Arguments */
    if      (nrhs!=26) { mexErrMsgIdAndTxt( "invalidNumInputs", "22 inputs required: C{t}, A{t}, Blhs{t}, Elhs{t}, ub{t}, lb{t}, x{t-1}, Meq(t), cutlhs, cutcoefs, settings.initialtheta, indices{t}, mexpdfs{t}, mexscales{t}, mexsigmas{t}, mexfunction, nburnin_mex, nrequired_mex, nmax, min_accept, min_s_in "); }
    else if (nlhs!=6)  { mexErrMsgIdAndTxt( "invalidNumInputs", "5 outputs required: samples, naccepted_mex, ntotal_mex, forcedout, cplextime");}    
    
    /* Declare Reference Variables for Loops */
    int                     i, j, s;
    
    /* Declare and Obtain Reference Variables for LP Size */
    int     N                   = (int) mxGetM(prhs[C_IN]);
    int     Mt                  = (int) mxGetM(prhs[A_IN]);
    int     *matlabMeq          = (int *) mxGetData(prhs[MEQ_IN]);
    int     Meq                 = *matlabMeq;
    int     Nx                  = (int) mxGetM(prhs[X_IN]);
    int     Mcuts               = (int) mxGetM(prhs[CUTLHS_IN]);
    int     Nt                  = N + (Mcuts > 0);
    int     D                   = (int) mxGetM(prhs[START_IN]);
    
    /* Create Pointers for LP Input Variables */
    double  *C              = mxGetPr(prhs[C_IN]);
    double  *A              = mxGetPr(prhs[A_IN]);
    double  *B              = mxGetPr(prhs[B_IN]);
    double  *E              = mxGetPr(prhs[E_IN]);
    double  *ub             = mxGetPr(prhs[UB_IN]);
    double  *lb             = mxGetPr(prhs[LB_IN]);
    double  *x              = mxGetPr(prhs[X_IN]);
    double  *cutlhs         = mxGetPr(prhs[CUTLHS_IN]);
    double  *cutcoefs       = mxGetPr(prhs[CUTCOEFS_IN]);
    double  *thetalb        = mxGetPr(prhs[THETALB_IN]);
    
    /* Create Pointers for PDF Input Variables */
    int     *indices        = (int *) mxGetData(prhs[INDICES_IN]);
    int     *pdfs           = (int *) mxGetData(prhs[PDF_IN]);
    double  *scales         = mxGetPr(prhs[SCALES_IN]);
    double  *sigmas         = mxGetPr(prhs[SIGMAS_IN]);
    double  *stepfactor     = mxGetPr(prhs[STEPFACTOR_IN]);
    double  *epsilon        = mxGetPr(prhs[EPSILON_IN]);
    
    double  *acceptprobs;
    double  *steps;
    double  *finalstep;
    double  *currentmean;
    double  *newmean;
    
    
    
    /* Create Pointers / Variables for Calling MATLAB for Steps */
    mxArray *pS_matlab          = mxCreateNumericMatrix((mwSize) 1, (mwSize) 1, mxINT32_CLASS, mxREAL);
    int     *S_matlab           = (int *) mxGetData(pS_matlab);
    double  *min_accept_rate    = mxGetPr(prhs[MIN_ACCEPT_IN]);
    int     *min_S_matlab       = (int *) mxGetData(prhs[MIN_S_IN]);
    mxArray *randFun            = (mxArray *) prhs[RANDFUN_IN];
    int     nRandIn             = 2;
    int     nRandOut            = 2;
    mxArray *randInputs[nRandIn];
    mxArray *randOutputs[nRandOut];

    /* Create Pointers / Variables for Calling MATLAB for Cholesky */
    mxArray *Cov_matlab         = mxCreateDoubleMatrix((mwSize) D, (mwSize) D, mxREAL);
    mxArray *LCov_matlab        = mxCreateDoubleMatrix((mwSize) D, (mwSize) D, mxREAL);
    double  *Cov                = mxGetPr(Cov_matlab);
    double  *LCov               = mxGetPr(LCov_matlab);
    mxArray *cholFun            = (mxArray *) prhs[CHOLFUN_IN];
    int     nCholIn             = 2;
    int     nCholOut            = 1;
    mxArray *cholInputs[nCholIn];
    mxArray *cholOutputs[nCholOut];


    
    /* Create Pointers / Initialize Variables for MCMC */
    double  *startpoint     = mxGetPr(prhs[START_IN]);
    int     *matlabburnin   = (int *) mxGetData(prhs[NBURNIN_IN]);
    int     *matlabrequired = (int *) mxGetData(prhs[NREQUIRED_IN]);
    int     *matlabadaptive = (int *) mxGetData(prhs[NADAPTIVE_IN]);
    int     *matlabnmax     = (int *) mxGetData(prhs[NMAX_IN]);
    int     nburnin         = *matlabburnin;
    int     nrequired       = *matlabrequired;
    int     nadaptive       = *matlabadaptive;
    int     nmax            = *matlabnmax;
    int     count           = 0;
    int     totalcount      = 0;
    int     naccepted       = 0;
    
    /* Initialize Number of Samples */
    int     S = (int) ((nrequired + nburnin) / *min_accept_rate);
    if (S < *min_S_matlab){ S = *min_S_matlab; }
    
    double currentvalue;
    double currentobjval;
    double currentprob;
    double *currentsample;
    double *currentpoint;
    
    double proposalvalue;
    double proposalobjval;
    double proposalprob;
    double *proposalsample;
    double *proposalpoint;
    
    /* Initialize Output Variables */
    plhs[ACC_SAMPLES_OUT]   = mxCreateDoubleMatrix(0, 0, mxREAL);                       
    mxSetM(plhs[ACC_SAMPLES_OUT], (mwSize) nrequired);                                 
    mxSetN(plhs[ACC_SAMPLES_OUT], (mwSize) D);
    mxSetData(plhs[ACC_SAMPLES_OUT], mxMalloc(nrequired * D * sizeof(double))); 
    
    plhs[ACC_COUNTS_OUT]    = mxCreateDoubleMatrix(0, 0, mxREAL);                        
    mxSetM(plhs[ACC_COUNTS_OUT], (mwSize) nrequired);                                  
    mxSetN(plhs[ACC_COUNTS_OUT], (mwSize) 1);
    mxSetData(plhs[ACC_COUNTS_OUT], mxMalloc(nrequired * sizeof(double)));
    
    plhs[NTOTAL_OUT]        = mxCreateDoubleScalar(0.0);
    plhs[NACCEPT_OUT]       = mxCreateDoubleScalar(1.0);
    plhs[FORCEDOUT_OUT]     = mxCreateDoubleScalar(1.0);
    plhs[TIME_OUT]          = mxCreateDoubleScalar(0.0);
    
    /* Create Pointers to Output Variables*/
    double *acceptedsamples = mxGetPr(plhs[ACC_SAMPLES_OUT]);
    double *acceptedcounts  = mxGetPr(plhs[ACC_COUNTS_OUT]);
    double *nacceptedout    = mxGetPr(plhs[NACCEPT_OUT]);
    double *ntotalout       = mxGetPr(plhs[NTOTAL_OUT]);
    double *forcedout       = mxGetPr(plhs[FORCEDOUT_OUT]);
    double *runtime         = mxGetPr(plhs[TIME_OUT]);
    double                  beforetime, aftertime;
    
    /*Initialize CPLEX Environment */
    CPXENVptr       env     = NULL;
    CPXLPptr        lp      = NULL;
    int             status  = 0;
    
    /*Declare Arrays Used to Build LPs in CPLEX */
    double          *lhs;
    double          *rmatval;
    int				*rmatind;
    int				*rmatbeg;
    char            *rmatsense;
    int				nnz = 0;
    
    /* Declare Arrays Used to Build Cut Variables in CPLEX */
    double          ctheta = 1, thetaub = CPX_INFBOUND;
    double          *cmatval;
    int				*cmatind;
    int				*cmatbeg;
    char            *cmatsense;
    int				nnzcuts = 0;
    
    /* Declare Variables and Pointers Used to Change Random LP Parameters*/
    int           s_obj_cnt = 0,    s_ub_cnt = 0,		 s_lb_cnt = 0,	 s_coef_cnt = 0;
    int          *s_obj_ind,        *s_ub_ind,			*s_lb_ind,		*s_coef_rows,    *s_coef_cols,    *s_lhs_ind;
    double       *s_obj_vals,       *s_ub_vals,         *s_lb_vals,		*s_coef_vals;
    char                            *s_ub_lu,			*s_lb_lu;
    
    /* Count Number of RVs in Each LP Parameter */
    for (i = 0; i < D; i ++ ){
        switch ( indices[i] ) {
            /*C*/   case 1:  s_obj_cnt++  ; break;
            /*A*/   case 2:  s_coef_cnt++ ; break;
            /*Aeq*/ case 3:  s_coef_cnt++ ; break;
            /*ub*/  case 8:  s_ub_cnt++   ; break;
            /*lb*/  case 9:  s_lb_cnt++   ; break;
        }
    }
    
    
    /* Flag LP  Parameters Types Containing RVs */
    int CHGOBJ  = ( s_obj_cnt > 0);
    int CHGCOEF = ( s_coef_cnt > 0);
    int CHGLHS  = ( D - s_obj_cnt -s_coef_cnt -s_ub_cnt -s_lb_cnt > 0 );
    int CHGUB   = ( s_ub_cnt > 0);
    int CHGLB   = ( s_lb_cnt > 0);
    
    
    /* Allocate Space and Initialize RV Variables for C	*/
    if ( CHGOBJ ) {
        s_obj_ind        = (int *)      malloc (s_obj_cnt   * sizeof(int));
        s_obj_vals       = (double *)   malloc (s_obj_cnt   * sizeof(double));
        
        j = 0;
        for (i = 0; i < D; i++) {
            if ( indices[i] == 1) {
                s_obj_ind[j] = indices[i+D];
                j ++;
            }
        }
    }
    
    /* Allocate Space and Initialize RV Variables for A and Aeq */
    if ( CHGCOEF ) {
        s_coef_vals      = (double *)   malloc (s_coef_cnt  * sizeof(double));
        s_coef_cols      = (int *)      malloc (s_coef_cnt  * sizeof(int));
        s_coef_rows      = (int *)      malloc (s_coef_cnt  * sizeof(int));
        
        j = 0;
        for (i = 0; i < D; i++) {
            if ( indices[i] == 2 || indices[i] == 3) {
                s_coef_rows[j] = (int) indices[i+D] % Mt;
                s_coef_cols[j] = (int) (indices[i+D] - s_coef_rows[j]) / Mt;
                j ++;
            }
        }
    }
    
    /* Allocate Space and Initialize RV Variables for lhs vector ( B, Beq, E, Eq) */
    if ( CHGLHS ) {
        s_lhs_ind          = (int *) malloc (Mt * sizeof(int));
        for (i = 0; i < Mt; i++) {
            s_lhs_ind[i] = i;
        }
    }
    
    /* Allocate Space and Initialize RV Variables for ub */
    if ( CHGUB ) {
        s_ub_ind         = (int *)      malloc (s_ub_cnt    * sizeof(int));
        s_ub_vals        = (double *)   malloc (s_ub_cnt    * sizeof(double));
        s_ub_lu          = (char *)     malloc (s_ub_cnt    * sizeof(char));
        
        j = 0;
        for (i = 0; i < D; i++) {
            if ( indices[i] == 1) {
                s_ub_ind[j] = indices[i+D];
                s_ub_lu[j]  = 'G';
                j ++;
            }
        }
    }
    
    /* Allocate Space and Initialize RV Variables for lb */
    if ( CHGLB ) {
        s_lb_ind         = (int *)      malloc (s_lb_cnt    * sizeof(int));
        s_lb_lu          = (char *)     malloc (s_lb_cnt    * sizeof(char));
        s_lb_vals        = (double *)   malloc (s_lb_cnt    * sizeof(double));
        
        j = 0;
        for (i = 0; i < D; i++) {
            if ( indices[i] == 1) {
                s_lb_ind[j] = indices[i+D];
                s_lb_lu[j]  = 'L';
                j ++;
            }
        }
    }
    
    
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    /*                                                                       */
    /*                           Setup LP Parameters                         */
    /*                                                                       */
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    
    /*Open CPLEX Environment */
    status = 1;
    while ( status ) {
        pthread_mutex_lock(&openCPLEXLock);
        env = CPXopenCPLEX (&status);
        pthread_mutex_unlock(&openCPLEXLock);
    }
    
    if ( env == NULL ) {
        mexPrintf("%s", "Could not open CPLEX environment.\n");
        goto TERMINATE;
    }
    
    
    /*Create CPLEX LP Object */
    lp = CPXcreateprob (env, &status, "LP");
    if ( lp == NULL ) {
        mexPrintf("%s", "Failed to create LP.\n");
        goto TERMINATE;
    }
    
    /* Allocate Memory According to Array Sizes */
    lhs             = (double *)    malloc (Mt          * sizeof(double));
    rmatval         = (double *)    malloc (Mt * N      * sizeof(double));
    rmatind         = (int *)       malloc (Mt * N      * sizeof(int));
    rmatbeg         = (int *)       malloc (Mt          * sizeof(int));
    rmatsense       = (char *)      malloc (Mt          * sizeof(char));
    
    /* Create Sense Vector for Constraints */
    for (i = 0;   i < Meq;  i++) { rmatsense[i] = 'E';} /*Equality Constraints*/
    for (i = Meq; i < Mt;   i++) { rmatsense[i] = 'G';} /*Inequality Constraints*/
    
    /* Add LP Objective */
    CPXchgobjsen (env, lp, CPX_MIN);
    status = CPXnewcols(env, lp, N, C, lb, ub, NULL, NULL);
    
    /* Add Equality and Inequality Constraints */
    for (i = 0; i < Mt; i++) {
        rmatbeg[i] = nnz;
        for (j = 0; j < N; j++) {
            rmatind[nnz] = j;
            rmatval[nnz] = A[i + j * Mt];
            nnz ++;
        }
    }
    
    /* Calculate Constraint LHS */
    for (i = 0; i < Mt; i++) {
        lhs[i] = B[i];
        for (j = 0; j < Nx; j++) {
            lhs[i] = lhs[i] - E[i + j * Mt] * x[j];
        }
    }
    
    status = CPXaddrows(env, lp, 0, Mt, nnz, lhs, rmatsense, rmatbeg, rmatind, rmatval, NULL, NULL);
    
    /* Free Memory */
    free(rmatbeg);
    free(rmatind);
    free(rmatval);
    free(rmatsense);
    
    /* Add Cut Variable and Cut Constraints */
    if (Mcuts > 0) {
        cmatsense       = (char *)   malloc (Mcuts          * sizeof(char));
        cmatval         = (double *) malloc (Mcuts * Nt     * sizeof(double));
        cmatind         = (int *)    malloc (Mcuts * Nt     * sizeof(int));
        cmatbeg         = (int *)    malloc (Mcuts			* sizeof(int));
        
        for (i = 0; i < Mcuts; i++)
        {
            cmatsense[i] = 'G';
            cmatbeg[i]   = nnzcuts;
            for (j = 0; j < Nt; j++) {
                cmatind[nnzcuts] = j;
                cmatval[nnzcuts] = cutcoefs[i + j * Mcuts];
                nnzcuts++;
            }
        }
        
        /* Add Cut Variable */
        status = CPXnewcols (env, lp, 1, &ctheta, thetalb, &thetaub, NULL, NULL);
        
        /* Add Cut Constraints */
        status = CPXaddrows (env, lp, 0, Mcuts, nnzcuts, cutlhs, cmatsense, cmatbeg, cmatind, cmatval, NULL, NULL);
        
        /* Free Memory */
        free(cmatbeg);
        free(cmatind);
        free(cmatval);
        free(cmatsense);
    }
    
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    /*                                                                       */
    /*                              MCMC                                     */
    /*                                                                       */
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    
    /* Allocate Memory for MCMC Variables */
    currentsample   = (double *) malloc (D  * sizeof(double));
    currentpoint    = (double *) malloc (D  * sizeof(double));
    currentmean     = (double *) malloc (D  * sizeof(double));
    proposalsample  = (double *) malloc (D  * sizeof(double));
    proposalpoint   = (double *) malloc (D  * sizeof(double));   
    newmean         = (double *) malloc (D  * sizeof(double));
    finalstep       = (double *) malloc (D  * sizeof(double));

    /* Set CPLEX Parameters */
    CPXsetintparam(env, CPX_PARAM_THREADS, 1);
    
    /* Initialize Current Values for MCMC */
    for (i = 0; i < D; i++){
        currentpoint[i]  = startpoint[i];
        currentsample[i] = strans(currentpoint[i], scales[i], pdfs[i]);
    }
    currentprob     = mvnpdf(currentpoint, sigmas, D);
    
    /* Set LP Parameters to Current Sample */
    j = 0;
    for (i = 0; i < D; i ++) {
        switch (indices[i]) {
            /*C*/   case 1: s_obj_vals[j]   = currentsample[i]; if (s_obj_cnt > j + 1){j++;}  else { j = 0;} break;
            /*A*/   case 2: s_coef_vals[j]  = currentsample[i]; if (s_coef_cnt > j + 1){j++;} else { j = 0;} break;
            /*Aeq*/ case 3: s_coef_vals[j]  = currentsample[i]; if (s_coef_cnt > j + 1){j++;} else { j = 0;} break;
            /*B*/   case 4: B[indices[i+D]] = currentsample[i];                                              break;
            /*Beq*/ case 5: B[indices[i+D]] = currentsample[i];                                              break;
            /*E*/   case 6: E[indices[i+D]] = currentsample[i];                                              break;
            /*Eeq*/ case 7: E[indices[i+D]] = currentsample[i];                                              break;
            /*ub*/  case 8: s_ub_vals[j]    = currentsample[i]; if (s_ub_cnt > j + 1){j++;}  else { j = 0;}  break;
            /*lb*/  case 9: s_lb_vals[j]    = currentsample[i]; if (s_lb_cnt > j + 1){j++;}  else { j = 0;}  break;
        }
    }
    
    /* Change LP Parameters in CPLEX LP to New Sample Values */
    if ( CHGOBJ ) { status = CPXchgobj(env, lp, s_obj_cnt, s_obj_ind, s_obj_vals);}
    if ( CHGCOEF ){ status = CPXchgcoeflist(env, lp, s_coef_cnt, s_coef_rows, s_coef_cols, s_coef_vals); }
    if ( CHGUB )  { status = CPXchgbds(env, lp, s_ub_cnt, s_ub_ind, s_ub_lu, s_ub_vals); }
    if ( CHGLB )  { status = CPXchgbds(env, lp, s_lb_cnt, s_lb_ind, s_lb_lu, s_lb_vals);}
    if ( CHGLHS ) {
        for (i = 0; i < Mt; i++) {
            lhs[i] = B[i];
            for (j = 0; j < Nx; j++) { lhs[i] = lhs[i] - E[i + j * Mt] * x[j]; }
        }
        status = CPXchgrhs(env, lp, Mt, s_lhs_ind, lhs);
    }
    
    /* Optimize LP and Obtain Current Objective Value */
    status = CPXgettime(env, &beforetime);
    status = CPXlpopt(env, lp);
    status = CPXgettime(env, &aftertime);
    *runtime = *runtime + (aftertime - beforetime);
    status = CPXgetobjval(env, lp, &currentobjval);
    
    /* Obtain Current Value */
    currentvalue = fabs(currentobjval) * currentprob;

    /*Initialize Covariance / L Covariance / and Current Mean*/
    for (i = 0; i < D; i++) {
        for (j = 0; j < D; j++) {
            Cov[i+D*j] = 0.0;
            LCov[i+D*j] = 0.0;
        }
        LCov[i+D*i]     = *stepfactor;
        currentmean[i]  = 0.0;
    }
    
    /*Initialize */
    cholInputs[0] = cholFun;
    cholInputs[1] = Cov_matlab;
    
    int burnmarker;
            
    /* Metropolis Hastings Loop */
    while ( (naccepted < nrequired) && (totalcount < nmax) ) {
        
        /*Obtain Acceptance Probs + Steps */
        S_matlab[0]     = S;
        randInputs[0]   = randFun;
        randInputs[1]   = pS_matlab;
        mexCallMATLAB(nRandOut, randOutputs, nRandIn, randInputs, "feval");
        acceptprobs     = mxGetPr(randOutputs[0]);
        steps           = mxGetPr(randOutputs[1]);

        for (s = 0; s < S; s++) {
            
            /* Obtain Final Step */
            for (i = 0; i < D; i++) {
                finalstep[i] = 0.0;
                for (j = 0; j <= i; j++){
                    finalstep[i] = finalstep[i] + LCov[i+j*D]*steps[s+j*S];
                }
            }

            /* Set Proposal Point / Sample / Probability */
            for (i = 0; i < D; i++){
                proposalpoint[i]  = currentpoint[i] + finalstep[i];
                proposalsample[i] = strans(proposalpoint[i], scales[i], pdfs[i]);
            }
            proposalprob   = mvnpdf(proposalpoint, sigmas, D);
            
            /* Set LP Parameters to Proposal Sample */
            j = 0;
            for (i = 0; i < D; i ++) {
                switch (indices[i]) {
                    /*C*/   case 1: s_obj_vals[j]   = proposalsample[i]; if (s_obj_cnt > j + 1){j++;}  else { j = 0;} break;
                    /*A*/   case 2: s_coef_vals[j]  = proposalsample[i]; if (s_coef_cnt > j + 1){j++;} else { j = 0;} break;
                    /*Aeq*/ case 3: s_coef_vals[j]  = proposalsample[i]; if (s_coef_cnt > j + 1){j++;} else { j = 0;} break;
                    /*B*/   case 4: B[indices[i+D]] = proposalsample[i];                                              break;
                    /*Beq*/ case 5: B[indices[i+D]] = proposalsample[i];                                              break;
                    /*E*/   case 6: E[indices[i+D]] = proposalsample[i];                                              break;
                    /*Eeq*/ case 7: E[indices[i+D]] = proposalsample[i];                                              break;
                    /*ub*/  case 8: s_ub_vals[j]    = proposalsample[i]; if (s_ub_cnt > j + 1){j++;}  else { j = 0;}  break;
                    /*lb*/  case 9: s_lb_vals[j]    = proposalsample[i]; if (s_lb_cnt > j + 1){j++;}  else { j = 0;}  break;
                }
            }
            
            /* Change LP Parameters in CPLEX LP to New Sample Values */
            if ( CHGOBJ ) { status = CPXchgobj(env, lp, s_obj_cnt, s_obj_ind, s_obj_vals);}
            if ( CHGCOEF ){ status = CPXchgcoeflist(env, lp, s_coef_cnt, s_coef_rows, s_coef_cols, s_coef_vals); }
            if ( CHGUB )  { status = CPXchgbds(env, lp, s_ub_cnt, s_ub_ind, s_ub_lu, s_ub_vals); }
            if ( CHGLB )  { status = CPXchgbds(env, lp, s_lb_cnt, s_lb_ind, s_lb_lu, s_lb_vals);}
            if ( CHGLHS ) {
                for (i = 0; i < Mt; i++) {
                    lhs[i] = B[i];
                    for (j = 0; j < Nx; j++) { lhs[i] = lhs[i] - E[i + j * Mt] * x[j]; }
                }
                status = CPXchgrhs(env, lp, Mt, s_lhs_ind, lhs);
            }
            
            /* Optimize LP and Obtain Proposal Objective Value */
            status = CPXgettime(env, &beforetime);
            status = CPXlpopt(env, lp);
            status = CPXgettime(env, &aftertime);
            *runtime = *runtime + (aftertime - beforetime);
            status = CPXgetobjval(env, lp, &proposalobjval);
            
            /*Calculate Proposal Value */
            proposalvalue = fabs(proposalobjval) * proposalprob;
            
            /*Accept / Reject */
            if ( proposalvalue > acceptprobs[s] * currentvalue ) {/*If Accepted */
                
                /* Set Current Value / Point to Proposal Sample / Point */
                currentvalue     =  proposalvalue;
                for (i = 0; i < D; i++){ currentpoint[i]     =  proposalpoint[i];}
                
                /* Update Count/Objval/Totals in Output after Burnin*/
                if (count >= nburnin ) {
                    
                    for (i = 0; i < D; i++){
                        acceptedsamples[naccepted + i*nrequired] = currentpoint[i];
                    }
                    
                    acceptedcounts[naccepted] = 1.0;
                    naccepted++;
                    if (naccepted == nrequired) { break;}
                }
                
                /*Increment all accepted count (including burnin) */
                if (count == nburnin){ burnmarker = s;}
                count++;
                
            } else if (count > nburnin ) {
                acceptedcounts[naccepted-1] = acceptedcounts[naccepted-1] + 1.0;   
            }
            
            /* Update Sample Covariance */
            if ( count >= nburnin ) {
                updateCovariance(Cov, currentpoint, currentmean, newmean, *stepfactor, *epsilon, D, s -burnmarker + 1);
            }
            
            /* Get Cholesky Decomposition of Cov -> LCov */
            if (naccepted >= nadaptive){
                mexCallMATLAB(nCholOut, cholOutputs, nCholIn, cholInputs, "feval");
                LCov = mxGetPr(cholOutputs[0]);
            }
        }
        
        /*Increment All Attempted Count */
        totalcount = totalcount + s + 1; 
        
        /*Calculate Required Samples */
        if (naccepted < nrequired){
            S = (int) ((double) nrequired - naccepted / ((double) count / totalcount));
        }
    }
    
    *nacceptedout = (double) naccepted;
    *ntotalout    = (double) totalcount;
    *forcedout    = (double) ( totalcount == nmax );
    
    goto TERMINATE;
    
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    /*                                                                       */
    /*                     Termination  Procedure                            */
    /*                                                                       */
    /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
    
    
    /* Termination Procedures */
    TERMINATE:
       
        /* Free up MCMC variables */
        if (proposalpoint != NULL)  { free(proposalpoint);  }
        if (proposalsample != NULL) { free(proposalsample); }
        if (currentpoint != NULL)   { free(currentpoint);   }
        if (currentsample != NULL)  { free(currentsample);  }
        if (currentmean != NULL)    { free(currentmean);    }
        if (newmean != NULL)        { free(newmean);        }
        if (finalstep != NULL)      { free(finalstep);      }
        if (lhs != NULL)            { free(lhs);            }
        
        /* Free up the problem as allocated by CPXcreateprob, if necessary */
        if ( lp != NULL ) {
            status = CPXfreeprob (env, &lp);
            if ( status ) { mexPrintf("%s%d\n", "CPXfreeprob failed, error code :", status); }
        }
        
        /* Free up the CPLEX environment, if necessary */
        if ( env != NULL ) {
            status = CPXcloseCPLEX (&env);
            if ( status ) {
                mexPrintf("%s\n", "Could not close CPLEX environment.");
            }
        }
        
        /* Free up variables used to change objective function */
        if ( CHGOBJ ) {
            if (s_obj_vals != NULL) { free(s_obj_vals);     }
            if (s_obj_ind != NULL)  { free(s_obj_ind);      }
        }
        
        if ( CHGUB ) {
            if (s_ub_lu != NULL)    { free(s_ub_lu);        }
            if (s_ub_ind != NULL)   { free(s_ub_ind);       } 
            if (s_ub_vals != NULL)  { free(s_ub_vals);      }
        }
        
        if ( CHGLB ) {
            if (s_lb_lu != NULL)    { free(s_lb_lu);        }
            if (s_lb_ind != NULL)   { free(s_lb_ind);       }
            if (s_lb_vals != NULL)  { free(s_lb_vals);      }
        }
        
        if ( CHGCOEF ) {
            if (s_coef_rows != NULL){ free(s_coef_rows);    }
            if (s_coef_cols != NULL){ free(s_coef_cols);    }
            if (s_coef_vals != NULL){ free(s_coef_vals);    }
        }
        
        if ( CHGLHS ) {
            if (s_lhs_ind != NULL)  { free(s_lhs_ind);      }
        }       
}